// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (SB16.CC) Copyright 1998 by Josef Drexler
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
// 
//     EMail:    bochs@world.std.com




#include "bochs.h"


#if BX_USE_SB16_SMF
bx_sb16_c bx_sb16;
#define this NULL
#endif

bx_sb16_c::bx_sb16_c(void)
{
  // nothing for now
}

bx_sb16_c::~bx_sb16_c(void)
{
  if (bx_options.sb16.loglevel > 0)
    fclose(LOGFILE);

  if (bx_options.sb16.midimode == 3)
    finishmidifile(MIDIDATA);
  if (bx_options.sb16.midimode > 0)
    fclose(MIDIDATA);

  if (bx_options.sb16.wavemode == 1)
    finishwavefile(WAVEDATA);
  if (bx_options.sb16.wavemode > 0)
    fclose(WAVEDATA);
}

  void
bx_sb16_c::init(bx_devices_c *d)
{
  BX_SB16_THIS devices = d;

  if ( (bx_options.sb16.logfile == NULL) ||
       (strlen(bx_options.sb16.logfile) < 1) )
    bx_options.sb16.loglevel = 0;

  if ( (bx_options.sb16.midifile == NULL) ||
       (strlen(bx_options.sb16.midifile) < 1) )
    bx_options.sb16.midimode = 0;

  if ( (bx_options.sb16.wavefile == NULL) ||
       (strlen(bx_options.sb16.wavefile) < 1) )
    bx_options.sb16.wavemode = 0;

  fprintf(stderr, "SB16 options: midi=%d,%s  wave=%d,%s  log=%d,%s\n",
	  bx_options.sb16.midimode, bx_options.sb16.midifile,
	  bx_options.sb16.wavemode, bx_options.sb16.wavefile,
	  bx_options.sb16.loglevel, bx_options.sb16.logfile);

  if (bx_options.sb16.loglevel > 0)
    LOGFILE = fopen(bx_options.sb16.logfile,"w"); // logfile for errors etc.

  if (bx_options.sb16.midimode > 0)
    {
      MIDIDATA = fopen(bx_options.sb16.midifile,"w"); // the midi stream is written to this file
      if (bx_options.sb16.midimode == 3)
	initmidifile(MIDIDATA);
    }

  if (bx_options.sb16.wavemode > 0)
    {
      WAVEDATA = fopen(bx_options.sb16.wavefile,"w"); // the wave stream
      if (bx_options.sb16.wavemode == 2)
	initwavefile(WAVEDATA);
    }

  MPUDATAIN init( (int) 64);	        // the input
  MPUDATAOUT init( (int) 64);	        // and output
  MPUCMD init( (int) 64);               // and command buffers
  MPUMIDICMD init( (int) 256);          // and the midi command buffer

  EMULIN init( (int) 64);               // the emulator ports
  EMULOUT init( (int) 64);              // for changing emulator settings

  BX_SB16_THIS mpu401.uartmode = 0;	// only generate a stream in uart mode

  writelog(3, "driver initialised, IRQ %d, IO %03x", 
	   BX_SB16_IRQ, BX_SB16_IO);

// the MPU IRQ is not implemented yet, and probably not even necessary
//  BX_SB16_THIS devices->register_irq(BX_SB16_IRQMPU, "SB16");
  BX_SB16_THIS devices->register_irq(BX_SB16_IRQ,"SB16");

// The SB16 takes many IO addresses
// First: 16 DSP/FM addresses
  for (unsigned addr=BX_SB16_IO; addr<=BX_SB16_IO+0xf; addr++) {
    BX_SB16_THIS devices->register_io_read_handler(this,
       read_handler,
       addr, "SB16");
    BX_SB16_THIS devices->register_io_write_handler(this,
       write_handler,
       addr, "SB16");
  }
    
// Second: 4 MPU addresses (status/command, data, unused, emulator ports)
  for (unsigned addr=BX_SB16_IOMPU; addr<=BX_SB16_IOMPU+3; addr++) {
    BX_SB16_THIS devices->register_io_read_handler(this,
       read_handler,
       addr, "SB16");
    BX_SB16_THIS devices->register_io_write_handler(this,
       write_handler,
       addr, "SB16");
    }

  if (bx_options.sb16.midimode > 1)
    {
      // initialize the timer
      BX_SB16_THIS timer_handle = bx_pc_system.register_timer
	                          ( & BX_SB16_THISU , midi_timer, 500000 / 384, 1, 1);
    }
  writelog(5, "Timer initialized, handle %d", BX_SB16_THIS timer_handle);
  BX_SB16_THIS current_timer = 0;
}

  // the timer functions
void bx_sb16_c::midi_timer (void *this_ptr)
{
  ((bx_sb16_c *) this_ptr)->current_timer++;
  writelog(5, "TIMER event - new timer: %d", ((bx_sb16_c *) this_ptr)->current_timer);
}

Bit32u bx_sb16_c::gettimer(void)
{
  return (BX_SB16_THIS current_timer);
}


  // the various IO handlers

  // First the DSP/FM music part

  // now the MPU 401 part


// the MPU 401 status port shows if input or output are ready
// Note that the bits are inverse to their meaning

  Bit32u
bx_sb16_c::mpu_status()
{
  Bit32u result = 0;

  if (MPUDATAIN full() == 1) result |= 0x40;	// no more output
  if (MPUDATAOUT empty() == 1) result |= 0x80;	// no input available

  writelog(4, "MPU status port, result %02x", result);
 
  return(result);
}


// the MPU 401 command port: Every byte is a command word, split into the two
// nibbles. More significant nibble is type of command, lower one
// selects the command
//
// Note: Except for Master Reset and Enter UART Mode, none of these are
//       supported at the moment. They will have to be implemented as
//       IOCTL calls to the /dev/midi00 driver, if it turns out to be necessary

  void
bx_sb16_c::mpu_command(Bit32u value)
{

  int success = 0;
  int bytesneeded;

  if (MPUCMD hascommand() == 1) // already a command pending, abort that one
    {
      if ( (MPUCMD currentcommand() != value) ||
	   (MPUCMD commanddone() == 0) )
	     // it's a different command, or the old one isn't done yet, abort it
	{
	  MPUCMD clearcommand();
	  MPUCMD flush();
	}
      // if it's the same one, and we just completed the argument list,
      // we leave it as it is and process it here
    }

  if (MPUCMD hascommand() == 0)  // no command pending, set one up
    {
      bytesneeded = 0;
      if ( (value >> 4) == 14) bytesneeded = 1;
      MPUCMD newcommand(value, bytesneeded);
    }

  if (MPUCMD commanddone() == 1) // command is complete, process it
    {
      switch (MPUCMD currentcommand() >> 4)
	{
/*	case 0:	switch (MPUCMD currentcommand() & 15) {	// MIDI commands
		case 1: writelog(5, "MIDI: Stop");
                        success=1;
			break;
		case 2: writelog(5, "MIDI: Start");
                        success=1;
			break;
		case 3: writelog(5, "MIDI: Continue");
                        success=1;
			break;
		}
		break;*/
/*	case 1:	switch (MPUCMD currentcommand() & 15) {	// system commands
		case 5:	writelog(5, "SYST: Stop device");
                        success=1;
 			break;
		}
		break;*/
	case 3: switch (MPUCMD currentcommand() & 15) {	// mode commands
/*		case 4: writelog(5, "MODE: Return timing bytes in stop mode");
                        success=1;
			break;
		case 5: writelog(5, "MODE: Enable mode messages");
                        success=1;
			break;
		case 8: writelog(5, "MODE: Enable system common messages");
                        success=1;
			break;
		case 9: writelog(5, "MODE: Enable realtime messages");
                        success=1;
			break;*/
		case 15:writelog(5, "MODE: UART mode on");
 			BX_SB16_THIS mpu401.uartmode=1;
	 		success=1;
 		}
	 	break;
/*	case 8:	switch (MPUCMD currentcommand() & 15) {	//  controller commands
		case 0:	writelog(5, "CTRL: Use internal MIDI timing");
                        success=1;
			break;
		case 1:	writelog(5, "CTRL: Use FSK timing");
                        success=1;
			break;
		case 2:	writelog(5, "CTRL: Use external MIDI timing");
                        success=1;
			break;
		case 3:	writelog(5, "CTRL: Enable metronome");
                        success=1;
			break;
		case 4:	writelog(5, "CTRL: Disable metronome");
                        success=1;
			break;
		case 7:	writelog(5, "CTRL: Enable pitch and controller");
                        success=1;
			break;
		case 10:writelog(5, "CTRL: Disable data in stop mode");
                        success=1;
			break;
		case 11:writelog(5, "CTRL: Enable data in stop mode");
                        success=1;
			break;
		}
		break;*/
/*	case 9:	switch (MPUCMD currentcommand() & 15) {	// Clock commands
		case 4:	writelog(5, "CLCK: Disable clock to host");
                        success=1;
			break;
		case 5: writelog(5, "CLCK: Enable clock to host");
                        success=1;
                        break;
		}
		break;*/
/*	case 10:switch (MPUCMD currentcommand() & 15) {	// Version commands
                case 12:writelog(5, "VERS: Return version number");
                        success=1;
                        break;
		case 13:writelog(5, "VERS: Return revision number");
			success=1;
			break;
		}
		break;*/
/*        case 12:  // Timebase commands
                writelog(5, "TBSE: Set timebase");
                success=1;
                break;*/
/*        case 14:switch (MPUCMD currentcommand() & 15) {   // Tempo commands
                case 0: writelog(5, "TEMP: Set tempo");
                        success=1;
                        break;
                case 4: writelog(5, "TEMP: Set metronome clocks per click");
                        success=1;
                        break;
                case 6: writelog(5, "TEMP: Set metronome beats per minutes");
                        success=1;
                        break;
                case 7: writelog(5, "TEMP: Send clocks to host");
                        success=1;
                        break;
                }
                break;*/
        case 15:switch (MPUCMD currentcommand() & 15) {   // Special commands
                case 15:writelog(5, "SPCL: Master reset of device");
			BX_SB16_THIS mpu401.uartmode=0;
                        success=1;
                        break;
                }
                break;
	}
      if (success == 0)
	writelog(3, "Unknown command");
      // Need to put an MPU_ACK into the data port if command successful
      // we'll fake it even if we didn't process the command, so as to 
      // allow detection of the MPU 401.
      if (MPUDATAOUT put(0xfe) == 0)
	writelog(3, "MPU_ACK error - output buffer full");
      MPUCMD clearcommand(); // clear the command from the buffer
    }

}


// MPU 401 data port/read: contains an MPU_ACK after receiving a command
// Will contain other data as well when other than UART mode is supported

  Bit32u
bx_sb16_c::mpu_dataread()
{
    Bit8u res8bit;
    Bit32u result;

    if ( MPUDATAOUT get(&res8bit) == 0)
      {
	writelog(3, "MPU data port not ready - no data in buffer");
	result = 0xff;
      }
    else
      result = (Bit32u) res8bit;

    writelog(4, "MPU data port, result %02x", result);

    return(result);
}


// MPU 401 data port/write: This is where the midi stream comes from,
// as well as arguments to any pending command

  void
bx_sb16_c::mpu_datawrite(Bit32u value)
{
  writelog(4, "write to MPU data port, value %02x", value);

  if (MPUCMD hascommand() == 1)
    { // there is a command pending, add arguments to it
      if (MPUCMD put(value) == 0)
	writelog(3, "MPU Command arguments too long - buffer full");
      if (MPUCMD commanddone() == 1)
	BX_SB16_THIS mpu_command(MPUCMD currentcommand());
    }
  else  // no MPU command pending, this has to be midi data
    {
      // first, find out if it is a midi command or midi data
      Boolean ismidicommand = 0;
      if (value >= 0x80)
	{  // bit 8 usually denotes a midi command
	  ismidicommand = 1;
	  if ( (value == 0xf7) && (MPUMIDICMD currentcommand() == 0xf0) )
	    // except if it is a continuing SYSEX message, then it just
	    // denotes the end of a SYSEX chunk
	    {
	      ismidicommand = 0;     // first, it's not a command
	      MPUMIDICMD newcommand(MPUMIDICMD currentcommand(), 
				    MPUMIDICMD bytes());
	                             // Then, set needed bytes to current buffer
	    }
	}
      
      if (ismidicommand == 1)
	{  // this is a command, check if an old one is pending
	  if (MPUMIDICMD hascommand() == 1)
	    {
	      writelog(3, "Midi command %d incomplete, has %d of %d bytes.", 
		       MPUMIDICMD currentcommand(), MPUMIDICMD bytes(), 
		       MPUMIDICMD commandbytes() );
	      // write as much as we can. Should we do this?
	      writemidicommand();
	      // clear the pending command
	      MPUMIDICMD clearcommand();
	      MPUMIDICMD flush();
	    }
	  
	  // find the number of arguments to the command
	  static signed eventlength[] = { 2, 2, 2, 2, 1, 1, 2, 255};
      	      // note - length 255 commands have unknown length
	  MPUMIDICMD newcommand(value, eventlength[ ( (value & 0xf0) >> 4 ) - 8 ]);
	}
      else  // no command, just arguments to the old command
	{
	  if (MPUMIDICMD hascommand() == 0)
	    {  // no command pending, ignore the data
	      writelog(3, "Midi data %d received, but no command pending?", value);
	      return;
	    }

	  // just some data to the command
	  if (MPUMIDICMD put(value) == 0)
	    writelog(3, "Midi buffer overflow!");
	  if (MPUMIDICMD commanddone() == 1)
	    {
	      // the command is complete, process it
	      writelog(5, "Midi command %d complete, has %d bytes.", 
		       MPUMIDICMD currentcommand(), MPUMIDICMD bytes() );
	      writemidicommand();
	      // and remove the command from the buffer
	      MPUMIDICMD clearcommand();
	      MPUMIDICMD flush();
	    }
	}
    }
}


// The emulator port/read: See if commands were successful

  Bit32u
bx_sb16_c::emul_read()
{
    Bit8u res8bit;
    Bit32u result;

    if ( EMULOUT get(&res8bit) == 0)
      {
	writelog(3, "emulator port not ready - no data in buffer");
	result = 0xff;
      }
    else
      result = (Bit32u) res8bit;

    writelog(4, "emulator port, result %02x", result);

    return(result);
}

// Emulator port/write: Changing instrument mapping etc.

  void
bx_sb16_c::emul_write(Bit32u value)
{

  Bit8u value8;

  writelog(4, "write to emulator port, value %02x", value);

  if (EMULOUT hascommand() == 0)   // no command pending, set it up
    {
      static signed char cmdlength[] = { 0, 0, 4, 2, 6 };      
      EMULOUT newcommand(value, 4);
      EMULIN put(0xfe);
    }
  else
    EMULOUT put(value);    // otherwise just add data

  if (EMULOUT commanddone() == 1)
    { // process the command
      switch (EMULOUT currentcommand())
	{
	case 0: // reinit of emulator
	        break;
	case 1: // dummy command to reset state of emulator port
                break;
	case 2: // map bank
	        break;
	case 3: // map program change
	        break;
	case 4: // map bank and program change
	        break;
	case 5: EMULOUT get(&value8); // dump emulator state
	        switch (value8)
		  {
		  case 0: EMULIN puts("SB16 Emulator for Bochs\n");
		          break;
		  case 1: EMULIN puts("UART mode=%d\n", BX_SB16_THIS mpu401.uartmode);
		          break;
		  case 2: EMULIN puts("timer=%d\n", BX_SB16_THIS current_timer);
		          break;
		  }
	        break;
	case 6: // close midi and wave files
	        if (bx_options.sb16.midimode == 3)
		  finishmidifile(MIDIDATA);
		if (bx_options.sb16.midimode > 0)
		  fclose(MIDIDATA);
		bx_options.sb16.midimode = 0;

		if (bx_options.sb16.wavemode == 1)
		  finishwavefile(WAVEDATA);
		if (bx_options.sb16.wavemode > 0)
		  fclose(WAVEDATA);	     
		bx_options.sb16.wavemode = 0;
		break;
	}
      EMULOUT clearcommand();
      EMULOUT flush();
    }
}


/* Handlers for the midi commands/midi file output */

// Write the header of the midi file. Track length is 0x7fffffff 
// until we know how long it's going to be

  void
bx_sb16_c::initmidifile(FILE *midifile)
{
  fwrite("MThd\000\000\000\006\000\001\000\001\001\200"
         "MTrk\177\377\377\177"
         "\000\377\121\003\007\241\140"
         "\000\377\130\004\004\002\030\010", 1, 37, midifile);
  return;
}


// write the midi command stored in MPUMIDICMD

  void
bx_sb16_c::writemidicommand()
{
  int i;
  Bit8u value;

  /* We need to determine the time elapsed since the last command */
  if (bx_options.sb16.midimode > 1)
    {
      Bit32u deltatime;
      static Bit32u oldtime = 0xffffffff;

      if (oldtime == 0xffffffff) oldtime = BX_SB16_THIS gettimer();

      /* Write delta time in special delta time format */
      deltatime = BX_SB16_THIS gettimer() - oldtime;
      writedeltatime(deltatime, MIDIDATA);

      /* Record the time */
      oldtime = BX_SB16_THIS gettimer();
    }

  // we need to log bank changes and program changes
  if ( (MPUMIDICMD currentcommand() >> 4) == 0xc)
    {    // a program change
      writelog(1, "* ProgramChange channel %d to %d",
	       MPUMIDICMD currentcommand() & 0xf,
	       MPUMIDICMD peek(0));
    }
  else if ( (MPUMIDICMD currentcommand() >> 4) == 0xb)
    {    // a control change, could be a bank change
      if (MPUMIDICMD peek(0) == 0)
	{  // bank select MSB
	  writelog(1, "* BankSelectMSB channel %d to %d",
		   MPUMIDICMD currentcommand() & 0xf,
		   MPUMIDICMD peek(1));
	}
      else if (MPUMIDICMD peek(0) == 32)
	{  // bank select LSB
	  writelog(1, "* BankSelectLSB channel %d to %d",
		   MPUMIDICMD currentcommand() & 0xf,
		   MPUMIDICMD peek(1));
	}
    }

  i = 0;
  fputc(MPUMIDICMD currentcommand(), MIDIDATA);
  if ( (MPUMIDICMD currentcommand() == 0xf0) || 
       (MPUMIDICMD currentcommand() == 0xf7) )       
    // write event length for sysex/meta events
    writedeltatime(MPUMIDICMD bytes(), MIDIDATA);
  
  while (MPUMIDICMD empty() == 0)
    {
      MPUMIDICMD get(&value);
      fputc(value, MIDIDATA);
    }

  fflush(MIDIDATA);       // to start playing immediately
}


// write a delta time coded value to the midi file

   void
bx_sb16_c::writedeltatime(Bit32u deltatime, FILE *midifile)
{
Bit8u outbytes[4];
int count = 0,i;

if (deltatime <= 0)
        fputc(0, midifile);
else {
        while ( (deltatime > 0) && (count < 4) ) {
                outbytes[count++] = (deltatime & 0x7f) | 0x80;
                deltatime >>= 7;
        };
        outbytes[0] &= 0x7f;
        for (i=count-1;i>=0;i--)
                fputc(outbytes[i], midifile);
}
}


// close the midi file, and set the track length accordingly

  void
bx_sb16_c::finishmidifile(FILE *midifile)
{
  fpos_t tracklen;

     // Meta event track end (0xff 0x2f 0x00) plus leading delta time
  fwrite("\000\377\057\000", 1, 4, midifile);

  fgetpos(midifile, &tracklen);
  tracklen -= 22;    // subtract the midi file header

  fseek(midifile, 22 - 4, SEEK_SET);

     // value has to be in big endian
#ifdef BX_LITTLE_ENDIAN
  tracklen = (tracklen << 24) | (tracklen >> 24) |
        ((tracklen & 0x00ff0000) >> 8) |
        ((tracklen & 0x0000ff00) << 8);
#endif
  fwrite(&tracklen, 4, 1, midifile);

  return;
}



/* Handlers for the wave file output */

// Write the header of the wave file. Track length is 0x7fffffff 
// until we know how long it's going to be

  void
bx_sb16_c::initwavefile(FILE *wavefile)
{
  // nothing yet
}


// close the wave file, and set the track length accordingly

  void
bx_sb16_c::finishwavefile(FILE *wavefile)
{
  // nothing yet
}



  // static IO port read callback handler
  // redirects to non-static class handler to avoid virtual functions

  Bit32u
bx_sb16_c::read_handler(void *this_ptr, Bit32u address, unsigned io_len)
{
#if !BX_USE_SB16_SMF
  bx_sb16_c *class_ptr = (bx_sb16_c *) this_ptr;

  return( class_ptr->read(address, io_len) );
}


  Bit32u
bx_sb16_c::read(Bit32u address, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_SB16_SMF
  UNUSED(address);
  UNUSED(io_len);

// we support only byte access to the port
  if (io_len != 1)
    {
      writelog(3, "not byte access!");
      return(0xff);
    }

  writelog(4, "access read at %03x for %d", address, io_len);

  if (address == BX_SB16_IOMPU+1)
      return(mpu_status());
  else if (address == BX_SB16_IOMPU)
      return(mpu_dataread());
  else if (address == BX_SB16_IOMPU+3)
      return(emul_read());

  // shouldn't get here
  writelog(3, "read from unsupported port!");

  return(0xff);
}

  // static IO port write callback handler
  // redirects to non-static class handler to avoid virtual functions

  void
bx_sb16_c::write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len)
{
#if !BX_USE_SB16_SMF
  bx_sb16_c *class_ptr = (bx_sb16_c *) this_ptr;

  class_ptr->write(address, value, io_len);
}

  void
bx_sb16_c::write(Bit32u address, Bit32u value, unsigned io_len)
{
#else
  UNUSED(this_ptr);
#endif  // !BX_USE_SB16_SMF

  writelog(4, "access write at %03x for %d to %02x", address, io_len, value);

  // we only support byte access to the prot
  if (io_len != 1)
	{
	  writelog(3, "not byte access!");
	  return;
	}

  if (address == BX_SB16_IOMPU)
      mpu_datawrite(value);
  else if (address == BX_SB16_IOMPU+1)
      mpu_command(value);
  else if (address == BX_SB16_IOMPU+3)
      emul_write(value);
  else
    writelog(3, "write to unsupported port!");

  return;
}

void bx_sb16_c::writelog(int loglevel, const char *str, ...)
{
// append a line to the log file, if desired
  if (bx_options.sb16.loglevel >= loglevel)
    {
	fprintf(LOGFILE, "SB16: (%i) ", loglevel);
	va_list ap;
	va_start(ap, str);
	vfprintf(LOGFILE, str, ap);
	va_end(ap);
	fprintf(LOGFILE, "\n");
    }
}

// the round-robin FIFO buffers of the SB16
bx_sb16_buffer::bx_sb16_buffer()
{
  length = 0;		// total bytes in buffer
  head = 0;		// pointer to next slot available for new data
  tail = 0;		// pointer to next slot to be read from
  buffer = NULL;	// pointer to the actual data
}

void bx_sb16_buffer::init(int bufferlen)
{
  if (buffer != NULL)	// Was it initialized before?
    delete buffer;

  length = bufferlen;
  buffer = new Bit8u[length];
  if (buffer == NULL)
    length = 0;		// This will be checked later
  head = 0;		// Reset the pointers
  tail = 0;

  clearcommand();       // no current command set
}

bx_sb16_buffer::~bx_sb16_buffer(void)
{
  if (buffer != NULL)
    delete buffer;
}


// Report how many bytes are available
int bx_sb16_buffer::bytes(void)
{
  int bytes = head - tail;
  if (bytes < 0) bytes += length;
  return (bytes);
}

// This puts one byte into the buffer
Boolean bx_sb16_buffer::put(Bit8u data)
{
  if (full() != 0)
    return 0;	// buffer full

  buffer[head++] = data;	// Write data, and increase write pointer
  head %= length;		// wrap it around so it stays inside the data

  return 1;	// put was successful
}

// This writes a formatted string to the buffer
Boolean bx_sb16_buffer::puts(char *data, ...)
{
  if (data == NULL)
    return 0;  // invalid string

  char string[length];
  int index = 0;

  va_list ap;
  va_start(ap, data);
  vsnprintf(string, length, data, ap);
  va_end(ap);
  
  while (string[index] != 0)
    {
      if (put( (Bit8u) string[index]) == 0)
	return 0;  // buffer full
      index++;
    }
  return 1;
}

// This returns if the buffer is full, i.e. if a put will fail
Boolean bx_sb16_buffer::full(void)
{
  if (length == 0)
    return 1;   // not initialized

  if ( ((head + 1) % length) == tail)
    return 1;   // buffer full

  return 0;	// buffer has some space left
}

// This reads the next available byte from the buffer
Boolean bx_sb16_buffer::get(Bit8u *data)
{
  if (empty() != 0)
    return 0;	// buffer empty

  (*data) = buffer[tail++];	// read data and increase read pointer
  tail %= length;		// and wrap it around to stay inside the data

  return 1;	// get was successful
}

// This returns if the buffer is empty, i.e. if a get will fail
Boolean bx_sb16_buffer::empty(void)
{
  if (length == 0)
    return 1;   // not inialized

  if (head == tail)
    return 1;   // buffer empty

  return 0;	// buffer contains data
}


// Flushes the buffer
void bx_sb16_buffer::flush(void)
{
  tail = head;
  return;
}

// Peeks ahead in the buffer
// Warning: No checking if result is valid. Must call bytes() to check that!
Bit8u bx_sb16_buffer::peek(int offset)
{
  return buffer[ (tail + offset) % length ];
}

// Set a new active command
void bx_sb16_buffer::newcommand(Bit8u newcmd, int bytes)
{
  command = newcmd;
  havecommand = 1;
  bytesneeded = bytes;
}

// Return the currently active command
Bit8u bx_sb16_buffer::currentcommand(void)
{
  return command;
}

// Clear the active command
void bx_sb16_buffer::clearcommand(void)
{
  command = 0;
  havecommand = 0;
  bytesneeded = 0;
}

// return if the command has received all necessary bytes
Boolean bx_sb16_buffer::commanddone(void)
{
  if (hascommand() == 0)
    return 0;  // no command pending - not done then

  if (bytes() >= bytesneeded)
    return 1;  // yes, it's done

  return 0;    // no, it's not
}

// return if there is a command pending
Boolean bx_sb16_buffer::hascommand(void)
{
  return havecommand;
}

int bx_sb16_buffer::commandbytes(void)
{
  return bytesneeded;
}
