// Copyright Notice
// ================
// BOCHS is Copyright 1994-1998 by Kevin P. Lawton.
//
// This file (SB16.H) Copyright 1998 by Josef Drexler
// 
// BOCHS is commercial software.
// 
// For more information, read the file 'LICENSE' included in the bochs
// distribution.  If you don't have access to this file, or have questions
// regarding the licensing policy, the author may be contacted via:
// 
//     US Mail:  Kevin Lawton
//               528 Lexington St.
//               Waltham, MA 02452-3028
// 
//     EMail:    bochs@world.std.com


#if BX_USE_SB16_SMF
#  define BX_SB16_SMF   static
#  define BX_SB16_THIS  bx_sb16.
#  define BX_SB16_THISU bx_sb16
#else
#  define BX_SB16_SMF
#  define BX_SB16_THIS  this->
#  define BX_SB16_THISU this
#endif

// some shortcuts to save typing
#define LOGFILE         BX_SB16_THIS logfile
#define MIDIDATA        BX_SB16_THIS midifile
#define WAVEDATA        BX_SB16_THIS wavefile
#define MPUDATAIN       BX_SB16_THIS mpu401.datain.
#define MPUDATAOUT      BX_SB16_THIS mpu401.dataout.
#define MPUCMD          BX_SB16_THIS mpu401.command.
#define MPUMIDICMD      BX_SB16_THIS mpu401.midicmd.
#define EMULIN          BX_SB16_THIS emuldata.datain.
#define EMULOUT         BX_SB16_THIS emuldata.dataout.

// the hardware properties. These can be changed via a DSP command
#define BX_SB16_IO      0x220
#define BX_SB16_IOMPU   0x330
#define BX_SB16_IRQ     5
#define BX_SB16_IRQMPU  9
#define BX_SB16_DMAL    1
#define BX_SB16_DMAH    5


// This is the class for the input and
// output FIFO buffers of the SB16
class bx_sb16_buffer {
public:

  bx_sb16_buffer(void);
  ~bx_sb16_buffer(void);
  void init(int bufferlen);

      /* These functions return 1 on success and 0 on error */
  Boolean put(Bit8u data);    // write on byte in the buffer
  Boolean puts(char *data, ...);  // write a formatted string to the buffer
  Boolean get(Bit8u *data);   // read the next available byte
  Boolean full(void);         // is the buffer full?
  Boolean empty(void);        // is it empty?

  void flush(void);           // empty the buffer
  int bytes(void);            // return number of bytes in the buffer
  Bit8u peek(int ahead);      // peek ahead number of bytes
 
      /* These are for caching the command number */
  void newcommand(Bit8u newcmd, int bytes);
  Bit8u currentcommand(void);
  void clearcommand(void);
  Boolean commanddone(void);
  Boolean hascommand(void);
  int commandbytes(void);


private:
  Bit8u *buffer;
  int head,tail,length;
  Bit8u command;
  Boolean havecommand;
  int bytesneeded;
  };

// The actual emulator class. It intercepts accesses to the
// SB16 IO ports and writes the raw midi stream to /dev/midi00
class bx_sb16_c {
public:

  bx_sb16_c(void);
  ~bx_sb16_c(void);
  BX_SB16_SMF void   init(bx_devices_c *);

private:

  FILE *logfile;
  FILE *midifile,*wavefile;
  struct {
    bx_sb16_buffer datain, dataout, command, midicmd;
    Boolean uartmode;
  } mpu401;

  struct {
    bx_sb16_buffer datain, dataout;
  } emuldata;

  int timer_handle;
  int current_timer;
  static void midi_timer (void *);
  static Bit32u gettimer (void);


  bx_devices_c *devices;

  static void   writelog(int loglevel, const char *str, ...);
  static void   writemidi(const Bit8u value);

      /* the MPU 401 part of the emulator */
  static Bit32u mpu_status();                   // read status port
  static void   mpu_command(Bit32u value);      // write command port
  static Bit32u mpu_dataread();                 // read data port
  static void   mpu_datawrite(Bit32u value);    // write data port
  static Bit32u emul_read (void);               // read emulator port
  static void   emul_write(Bit32u value);       // write emulator port

  static void   writemidicommand(void);         // flush the midi buffer
  static void   writedeltatime(Bit32u deltatime, FILE *midifile);
                                                // write in delta time coding

  static void   initmidifile(FILE *midifile);   // Write midi file header
  static void   finishmidifile(FILE *midifile); // write track length etc.
  static void   initwavefile(FILE *wavefile);   // Write wave file header
  static void   finishwavefile(FILE *wavefile); // write track length etc.

  //  static Bit32u dsp_read(void *this_ptr,  Bit32u address, unsigned io_len);
  //  static void   dsp_write(void *this_ptr,  Bit32u address, Bit32u value, unsigned io_len);

  static Bit32u read_handler(void *this_ptr, Bit32u address, unsigned io_len);
  static void   write_handler(void *this_ptr, Bit32u address, Bit32u value, unsigned io_len);
#if !BX_USE_SB16_SMF
  Bit32u read(Bit32u address, unsigned io_len);
  void   write(Bit32u address, Bit32u value, unsigned io_len);
#endif
  };


#if BX_USE_SB16_SMF
extern bx_sb16_c bx_sb16;
#endif
